
{% for  result in result_list %}
<details>
{{environment.get_template("md_stats_template.jinja").render(name = result.name, stats = result.stats)}}
<ul>
{{result.render()}}
</ul>
</details>
{% endfor %}